/* extension.js
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

/* exported init */

const Gettext = imports.gettext;
const Lang = imports.lang;

const { Clutter, GObject, St } = imports.gi;

const ExtensionUtils = imports.misc.extensionUtils;
const Util = imports.misc.util;

const Main = imports.ui.main;
const PanelMenu = imports.ui.panelMenu;
const PopupMenu = imports.ui.popupMenu;

const Me = ExtensionUtils.getCurrentExtension();

const Sensors = Me.imports.sensors;

const _ = Gettext.domain(Me.metadata['gettext-domain']).gettext;

const KEY_REFRESH_INTERVAL = 'refresh-interval';
const KEY_SHOW_FAHRENHEIT = 'show-fahrenheit';
const KEY_SHOW_TEMPERATURE = 'show-temperature';

const TEMPERATURE_AVERAGE = 'average';
const TEMPERATURE_DISABLED = '';
const TEMPERATURE_HIGHEST = 'highest';

function getTemperatureAsString(temperature) {
    if (!temperature || temperature === 0) return 'N/A';
    if (ExtensionUtils.getSettings().get_boolean(KEY_SHOW_FAHRENHEIT)) {
        return Math.round(temperature * 1.8 + 32).toFixed(0)+' \u00b0F';
    } else {
        return Math.round(temperature).toFixed(1)+' \u00b0C';
    }
}

const PopupTemperatureMenuItem = GObject.registerClass(
class PopupTemperatureMenuItem extends PopupMenu.PopupBaseMenuItem {
    _init(text, icon, params) {
        super._init(params);
        if (icon) {
            this.icon = new St.Icon({ icon_name: icon, style_class: 'popup-menu-icon', x_align: Clutter.ActorAlign.END });
            this.add_child(this.icon);
        }
        this.label = new St.Label({ text: text, x_expand: true, y_expand: true, y_align: Clutter.ActorAlign.CENTER });
        this.add_child(this.label);
        this.label_actor = this.label;
        this._statusLabel = new St.Label({ text: '', y_align: Clutter.ActorAlign.CENTER });
        this.add_child(this._statusLabel);
        this._settings = ExtensionUtils.getSettings();
        this._settingsChangedId = [
            this._settings.connect('changed::%s'.format(KEY_SHOW_FAHRENHEIT), this._onChangedShowFahrenheit.bind(this)),
            this._settings.connect('changed::%s'.format(KEY_SHOW_TEMPERATURE), this._onChangedShowTemperature.bind(this)),
        ];
    }

    _onChangedShowFahrenheit() {
        if (!this._statusLabel) return;
        this._statusLabel.set_text(getTemperatureAsString(this._status));
    }

    _onChangedShowTemperature() {
        if (!this._id) return;
        let showTemperature = this._settings.get_strv(KEY_SHOW_TEMPERATURE);
        this._selected = showTemperature.indexOf(this._id) >= 0;
        this.setOrnament(this._selected ? PopupMenu.Ornament.CHECK : PopupMenu.Ornament.NONE);
    }

    activate(event) {
        if (this._id) this._settings.set_strv(KEY_SHOW_TEMPERATURE, [ this._selected ? TEMPERATURE_DISABLED : this._id ]);
        super.activate(event);
    }

    destroy() {
        this._settingsChangedId.forEach(id => {
            this._settings.disconnect(id);
        });
        super.destroy();
    }

    setId(id) {
        this._id = id;
        this._onChangedShowTemperature();
    }

    setStatus(status) {
        this._status = status;
        this._onChangedShowFahrenheit();
    }
});

const PopupTemperatureSubMenuMenuItem = GObject.registerClass(
class PopupTemperatureSubMenuMenuItem extends PopupMenu.PopupSubMenuMenuItem {
    _init(text, icon) {
        super._init(text, true);
        this.icon.icon_name = icon;
        this._statusLabel = new St.Label({ y_expand: true, y_align: Clutter.ActorAlign.CENTER });
        let children = this.actor.get_n_children();
        this.actor.insert_child_at_index(this._statusLabel, children - 1);
        this._settings = ExtensionUtils.getSettings();
        this._settingsChangedId = this._settings.connect('changed::%s'.format(KEY_SHOW_FAHRENHEIT), this._onChangedShowFahrenheit.bind(this));
    }

    _onChangedShowFahrenheit() {
        this._statusLabel.set_text(getTemperatureAsString(this._status));
    }

    addMenuItem(text, status) {
        let item = new PopupTemperatureMenuItem(text, null, { reactive: false });
        item.setStatus(status);
        this.menu.addMenuItem(item);
    }

    destroy() {
        this._settings.disconnect(this._settingsChangedId);
        super.destroy();
    }

    setStatus(status) {
        this._status = status;
        this._onChangedShowFahrenheit();
    }
});

const PopupTemperatureSensorSubMenuMenuItem = GObject.registerClass(
class PopupTemperatureSensorSubMenuMenuItem extends PopupTemperatureSubMenuMenuItem {
    _init(sensor, text, icon) {
        super._init(text, icon)
        this._sensor = sensor;
        this._sensorChangedId = this._sensor.connect('updated::sensor', this._onUpdatedSensor.bind(this));
        this._onUpdatedSensor();
    }

    _onUpdatedSensor() {
        this.menu.removeAll();
        this._sensor.children.forEach(child => {
            this.addMenuItem(child.label, child.current);
        });
        this.setStatus(this._sensor.highest);
        this.actor.visible = this.menu.numMenuItems > 0;
    }

    destroy() {
        this._sensor.disconnect(this._sensorChangedId);
        super.destroy();
    }
});

const PopupTemperatureSensorMenuSection = class extends PopupMenu.PopupMenuSection {
    constructor(sensor) {
        super();
        this._sensor = sensor;
        this._sensorChangedId = this._sensor.connect('updated::sensor', this._onUpdatedSensor.bind(this));
        this._onUpdatedSensor();
    }

    _onUpdatedSensor() {
        this.removeAll();
        this._sensor.children.forEach(child => {
            let item;
            if (child.maximum || child.critical) {
                item = new PopupTemperatureSubMenuMenuItem(child.label, child.warning ? 'dialog-warning' : 'utilities-system-monitor');
                item.setStatus(child.current);
                item.addMenuItem(_("Current temperature"), child.current);
                if (child.maximum) item.addMenuItem(_("Maximum temperature"), child.maximum);
                if (child.critical) item.addMenuItem(_("Critical temperature"), child.critical);
            } else {
                item = new PopupTemperatureMenuItem(child.label, 'utilities-system-monitor', { reactive: false });
                item.setStatus(child.current);
            }
            this.addMenuItem(item);
        });
        this.actor.visible = this.numMenuItems > 0;
    }

    destroy() {
        this._sensor.disconnect(this._sensorChangedId);
        super.destroy();
    }
}

const TemperatureMenuButton = GObject.registerClass(
class TemperatureMenuButton extends PanelMenu.Button {
    _init() {
        super._init(0.5, _("Temperature"), false);
        this.menu.actor.add_style_class_name('dschaerf-menu');
        this._hbox = new St.BoxLayout({ style_class: 'panel-status-indicators-box' });
        this._icon = new St.Icon({ icon_name: 'view-refresh-symbolic', style_class: 'system-status-icon'});
        this._hbox.add_child(this._icon);
        this._label = new St.Label({ text: '...', y_expand: true, y_align: Clutter.ActorAlign.CENTER }),
        this._hbox.add_child(this._label);
        this._hbox.add_child(PopupMenu.arrowIcon(St.Side.BOTTOM));
        this.add_child(this._hbox);
        this._sensorManager = new Sensors.SensorManager();
        this._itemNotAvailable = new PopupTemperatureMenuItem(_("No supported temperature sensors available"), 'dialog-warning', { reactive: false, can_focus: false });
        this.menu.addMenuItem(this._itemNotAvailable);
        this._itemSensorHWM = new PopupTemperatureSensorMenuSection(this._sensorManager.sensor['hwm']);
        this.menu.addMenuItem(this._itemSensorHWM);
        this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
        this._itemSensorVGA = new PopupTemperatureSensorSubMenuMenuItem(this._sensorManager.sensor['vga'], _("Graphics Cards"), 'video-display');
        this.menu.addMenuItem(this._itemSensorVGA);
        this._itemSensorHDD = new PopupTemperatureSensorSubMenuMenuItem(this._sensorManager.sensor['hdd'], _("Hard disk drives"), 'drive-harddisk');
        this.menu.addMenuItem(this._itemSensorHDD);
        this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
        this._itemAverage = new PopupTemperatureMenuItem(_("Average temperature"), 'utilities-system-monitor');
        this._itemAverage.setId('average');
        this.menu.addMenuItem(this._itemAverage);
        this._itemHighest = new PopupTemperatureMenuItem(_("Highest temperature"), 'utilities-system-monitor');
        this._itemHighest.setId('highest');
        this.menu.addMenuItem(this._itemHighest);
        this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
        this._itemActions = new PopupMenu.PopupBaseMenuItem({ reactive: false, style_class: 'popup-menu-item', can_focus: false });
        this._actionSettings = this._createActionButton('preferences-other-symbolic');
        this._actionSettings.connect('clicked', this._onClickedSettings.bind(this));
        this._itemActions.add_child(this._actionSettings);
        let expander = new St.Bin({ style_class: 'popup-menu-item-expander', x_expand: true });
        this._itemActions.add_child(expander);
        this._actionRefresh = this._createActionButton('view-refresh-symbolic');
        this._actionRefresh.connect('clicked', this._onChangedRefreshInterval.bind(this));
        this._itemActions.add_child(this._actionRefresh);
        this.menu.addMenuItem(this._itemActions);
        this._settings = ExtensionUtils.getSettings();
        this._settingsChangedId = [
            this._settings.connect('changed::%s'.format(KEY_REFRESH_INTERVAL), this._onChangedRefreshInterval.bind(this)),
            this._settings.connect('changed::%s'.format(KEY_SHOW_FAHRENHEIT), this._updateLabel.bind(this)),
            this._settings.connect('changed::%s'.format(KEY_SHOW_TEMPERATURE), this._updateLabel.bind(this)),
        ];
        this._sensorManagerChangedId = this._sensorManager.connect('updated::sensors', this._onUpdatedSensors.bind(this));
        this._onChangedRefreshInterval();
    }

    _createActionButton(icon) {
        let button = new St.Button({ style_class: 'button action-button' });
        button.child = new St.Icon({ icon_name: icon, style_class: 'popup-menu-icon' });
        return button;
    }

    _onChangedRefreshInterval() {
        let refreshInterval = this._settings.get_string(KEY_REFRESH_INTERVAL);
        this._sensorManager.setRefreshInterval(parseInt(refreshInterval));
    }

    _onClickedSettings() {
        this.menu.itemActivated();
        Main.overview.hide();
        Util.spawnCommandLine('gnome-shell-extension-prefs '+Me.metadata['uuid']);
    }

    _onDestroy() {
        this._sensorManager.disconnect(this._sensorManagerChangedId);
        this._sensorManager.destroy();
        this._settingsChangedId.forEach(id => {
            this._settings.disconnect(id);
        });
        super._onDestroy();
    }

    _onOpenStateChanged(menu, open) {
        if (open) {
            let available = this._sensorManager.available;
            this._itemNotAvailable.actor.visible = !available;
            this._itemAverage.actor.visible = available;
            this._itemHighest.actor.visible = available;
            this._itemActions.actor.visible = available;
        }
        super._onOpenStateChanged(menu, open);
    }

    _onUpdatedSensors() {
        this._itemAverage.setStatus(this._sensorManager.average);
        this._itemHighest.setStatus(this._sensorManager.highest);
        this._updateIcon();
        this._updateLabel();
    }

    _updateIcon() {
        let icon = 'org.gnome.SystemMonitor-symbolic';
        if (!this._sensorManager.available || this._sensorManager.warning) icon = 'dialog-warning-symbolic';
        this._icon.icon_name = icon;
    }

    _updateLabel() {
        let showTemperature = this._settings.get_strv(KEY_SHOW_TEMPERATURE)[0];
        let visible = showTemperature !== TEMPERATURE_DISABLED;
        if (visible) {
            let temperature = this._sensorManager[showTemperature];
            this._label.set_text(getTemperatureAsString(temperature));
        }
        this._label.visible = visible;
    }
});

class Extension {
    constructor(uuid) {
        this._uuid = uuid;
        ExtensionUtils.initTranslations();
    }

    enable() {
        this._indicator = new TemperatureMenuButton();
        let pos = Main.sessionMode.panel.left.indexOf('dateMenu');
        Main.panel.addToStatusArea(this._uuid, this._indicator, pos, 'center');
    }

    disable() {
        this._indicator.destroy();
        this._indicator = null;
    }
}

function init(meta) {
    return new Extension(meta.uuid);
}
